'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var html = require('../html');

const isBrowser = typeof window !== "undefined";
const supportsAdoptedStyleSheets = isBrowser &&
    "adoptedStyleSheets" in Document.prototype &&
    "replace" in CSSStyleSheet.prototype;

const cssSymbol = Symbol("css");
const hasCSSSymbol = (value) => {
    return value && value[cssSymbol] != null;
};
const resolve = (value) => {
    if (typeof value === "number")
        return String(value);
    if (hasCSSSymbol(value))
        return value[cssSymbol];
    if (value instanceof URL)
        return value.href;
    throw new TypeError(`${value} is not supported type.`);
};
const css = (strings, ...values) => ({
    [cssSymbol]: strings
        .slice(1)
        .reduce((acc, s, i) => acc + resolve(values[i]) + s, strings[0])
});
const unsafeCSS = (css) => ({ [cssSymbol]: css });
const adoptCssStyle = (root, css) => {
    const cssStyle = css[cssSymbol];
    if (supportsAdoptedStyleSheets) {
        const styleSheet = new CSSStyleSheet();
        styleSheet.replace(cssStyle);
        root.adoptedStyleSheets = [...root.adoptedStyleSheets, styleSheet];
    }
    else {
        root.appendChild(document.createElement("style")).textContent = cssStyle;
    }
};

let currentCursor;
let currentComponent;
function defaultHooks() {
    return {
        _values: [],
        _deps: [],
        _effects: [],
        _layoutEffects: [],
        _cleanup: []
    };
}
function __setCurrent__(c) {
    currentComponent = c;
    currentCursor = 0;
}
function hooks(config) {
    const h = currentComponent._hooks;
    const c = currentComponent;
    const index = currentCursor++;
    if (h._values.length <= index && config._onmount) {
        h._values[index] = config._onmount(h, c, index);
    }
    if (config._onupdate) {
        config._onupdate(h, c, index);
    }
    return h._values[index];
}

const batch = (queue, pick, callback) => {
    const q = [];
    const flush = () => {
        let p;
        while ((p = pick(q)))
            callback(p);
    };
    return (c) => q.push(c) === 1 && queue(flush);
};
const fifo = (q) => q.shift();
const lifo = (q) => q.pop();
const queueTask = (callback) => {
    try {
        const ch = new MessageChannel();
        ch.port1.onmessage = callback;
        ch.port2.postMessage(null);
    }
    catch (_) {
        /* istanbul ignore next */
        setImmediate(callback);
    }
};
const enqueueLayoutEffects = batch(queueMicrotask, lifo, c => c._flushEffects(c._hooks._layoutEffects));
const enqueueEffects = batch(queueTask, lifo, c => c._flushEffects(c._hooks._effects));
const enqueueUpdate = batch(queueMicrotask, fifo, c => c._performUpdate());

if (!isBrowser) {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    global.HTMLElement = Function;
}
class Component extends HTMLElement {
    constructor() {
        super(...arguments);
        this._dirty = false;
        this._connected = false;
        this._container = this.attachShadow({ mode: "open" });
        this._hooks = defaultHooks();
    }
    connectedCallback() {
        this._connected = true;
        this.update();
    }
    disconnectedCallback() {
        this._connected = false;
        const cleanups = this._hooks._cleanup;
        for (let i = 0; i < cleanups.length; i++) {
            if (cleanups[i]) {
                cleanups[i]();
                delete cleanups[i];
            }
        }
    }
    update() {
        if (this._dirty)
            return;
        this._dirty = true;
        enqueueUpdate(this);
    }
    _performUpdate() {
        if (!this._connected)
            return;
        try {
            __setCurrent__(this);
            this.renderer(this._container);
            enqueueLayoutEffects(this);
            enqueueEffects(this);
        }
        catch (e) {
            console.error(e);
        }
        this._dirty = false;
    }
    _flushEffects(effects) {
        const cleanups = this._hooks._cleanup;
        for (let i = 0, len = effects.length; i < len; i++) {
            if (effects[i]) {
                cleanups[i] && cleanups[i]();
                const cleanup = effects[i]();
                if (typeof cleanup === "function") {
                    cleanups[i] = cleanup;
                }
                delete effects[i];
            }
        }
    }
    _attr(name, converter) {
        return converter
            ? converter(this.getAttribute(name))
            : this.getAttribute(name);
    }
    _observeAttr(name, callback) {
        const m = new MutationObserver(callback);
        m.observe(this, { attributes: true, attributeFilter: [name] });
        return () => m.disconnect();
    }
    _dispatch(name, init) {
        this.dispatchEvent(new CustomEvent(name, init));
    }
    _adoptStyle(css) {
        adoptCssStyle(this._container, css);
    }
}

const __FucoRegistry__ = {};
function defineElement(name, fn) {
    if (isBrowser) {
        customElements.define(name, class extends Component {
            renderer(container) {
                html.render(fn(), container);
            }
        });
    }
    else {
        __FucoRegistry__[name] = fn;
    }
}

const REQUEST_CONSUME = "context:request-consume:" + String(Math.random()).slice(2);

function useAttribute(name, converter) {
    return hooks({
        _onmount(h, c, i) {
            h._cleanup[i] = c._observeAttr(name, () => {
                const newValue = c._attr(name, converter);
                if (!Object.is(h._values[i], newValue)) {
                    h._values[i] = newValue;
                    c._performUpdate();
                }
            });
            return c._attr(name, converter);
        }
    });
}
const useProperty = (name) => hooks({
    _onmount(h, c, i) {
        const initialValue = c[name];
        Object.defineProperty(c, name, {
            get() {
                return h._values[i];
            },
            set(newValue) {
                if (!Object.is(h._values[i], newValue)) {
                    h._values[i] = newValue;
                    c.update();
                }
            }
        });
        return initialValue;
    }
});
const useDispatchEvent = (name, eventInit = {}) => hooks({
    _onmount: (_, c) => (detail) => c._dispatch(name, Object.assign(Object.assign({}, eventInit), { detail }))
});
const useStyle = (cssStyle) => hooks({
    _onmount(h, c, i) {
        h._layoutEffects[i] = () => c._adoptStyle(typeof cssStyle === "function" ? cssStyle() : cssStyle);
    }
});
const useRef = (initialValue) => hooks({
    _onmount: (_h, _c) => ({ current: initialValue })
});
const useState = (initialState) => hooks({
    _onmount: (h, c, i) => [
        typeof initialState === "function"
            ? initialState()
            : initialState,
        function setState(nextState) {
            const state = h._values[i][0];
            if (typeof nextState === "function") {
                nextState = nextState(state);
            }
            if (!Object.is(state, nextState)) {
                h._values[i][0] = nextState;
                c.update();
            }
        }
    ]
});
const useReducer = (reducer, initialState) => hooks({
    _onmount: (h, c, i) => [
        initialState,
        function dispatch(action) {
            const state = h._values[i][0];
            const nextState = reducer(state, action);
            if (!Object.is(state, nextState)) {
                h._values[i][0] = nextState;
                c.update();
            }
        }
    ]
});
const useContext = (context) => hooks({
    _onmount: (h, c, i) => {
        h._values[i] = context._defaultValue;
        c._dispatch(REQUEST_CONSUME, {
            bubbles: true,
            composed: true,
            detail: {
                _context: context,
                _register(subscribe, initialValue) {
                    h._values[i] = initialValue;
                    h._cleanup[i] = subscribe(nextValue => {
                        if (!Object.is(h._values[i], nextValue)) {
                            h._values[i] = nextValue;
                            c.update();
                        }
                    });
                }
            }
        });
        return h._values[i];
    }
});
const depsChanged = (prev, next) => !prev || next.some((f, i) => !Object.is(f, prev[i]));
const useEffect = (handler, deps) => hooks({
    _onupdate(h, _, i) {
        if (!deps || depsChanged(h._deps[i], deps)) {
            h._deps[i] = deps || [];
            h._effects[i] = handler;
        }
    }
});
const useLayoutEffect = (handler, deps) => hooks({
    _onupdate(h, _, i) {
        if (!deps || depsChanged(h._deps[i], deps)) {
            h._deps[i] = deps || [];
            h._layoutEffects[i] = handler;
        }
    }
});
const useMemo = (fn, deps) => hooks({
    _onupdate(h, _, i) {
        if (!deps || depsChanged(h._deps[i], deps)) {
            h._deps[i] = deps || [];
            h._values[i] = fn();
        }
    }
});
const useCallback = (callback, deps) => useMemo(() => callback, deps);

// import html as h to avoid format by prettier
function createContext(defaultValue) {
    const context = {
        _defaultValue: defaultValue,
        Provider() {
            const subs = useRef([]);
            const subscribe = (s) => {
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                const len = subs.current.push(s);
                return () => {
                    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                    subs.current.splice(len - 1, 1);
                };
            };
            const value = useProperty("value");
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            useEffect(() => subs.current.forEach(f => f(value)), [value]);
            const request = (e) => {
                if (context === e.detail._context) {
                    e.stopPropagation();
                    e.detail._register(subscribe, value);
                }
            };
            return html.html `<slot ...${{ [`@${REQUEST_CONSUME}`]: request }}/>`;
        },
        defineProvider(name) {
            defineElement(name, context.Provider);
        }
    };
    return context;
}

Object.defineProperty(exports, 'html', {
  enumerable: true,
  get: function () {
    return html.html;
  }
});
Object.defineProperty(exports, 'render', {
  enumerable: true,
  get: function () {
    return html.render;
  }
});
exports.Component = Component;
exports.__FucoRegistry__ = __FucoRegistry__;
exports.__setCurrent__ = __setCurrent__;
exports.createContext = createContext;
exports.css = css;
exports.defaultHooks = defaultHooks;
exports.defineElement = defineElement;
exports.unsafeCSS = unsafeCSS;
exports.useAttribute = useAttribute;
exports.useCallback = useCallback;
exports.useContext = useContext;
exports.useDispatchEvent = useDispatchEvent;
exports.useEffect = useEffect;
exports.useLayoutEffect = useLayoutEffect;
exports.useMemo = useMemo;
exports.useProperty = useProperty;
exports.useReducer = useReducer;
exports.useRef = useRef;
exports.useState = useState;
exports.useStyle = useStyle;
