'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

/* eslint-disable @typescript-eslint/no-non-null-assertion */
const openTagRegexp = /^\s*<\s*([a-zA-Z1-9-]+)/;
const closeTagRegexp = /^<\s*\/\s*([a-zA-Z1-9-]+)>/;
const tagEndRegexp = /^\s*(\/)?>/;
const voidTagNameRegexp = /^(area|base|br|col|embed|hr|img|input|link|meta|param|source|track|wbr)$/i;
const attributeNameRegexp = /^\s*([^\s"'<>\/=]+)(?:\s*(=))?/;
const attributeValueRegexp = /^\s*(?:\s*(?:"([^"]*)"+|'([^']*)'+|([^\s"'=<>`]+)))?/;
const doctypeRegexp = /^\s*<!DOCTYPE [^>]+>/i;
const commentStartRegexp = /^\s*<!--/;
const commentEndRegexp = /-->/;
function parse(htmls) {
    const root = { children: [] };
    const stack = [];
    let inTag = false;
    let inComment = false;
    let current = root;
    let attrName = "";
    function commit(value) {
        if (inComment)
            return;
        /* istanbul ignore else */
        if (inTag && attrName) {
            (current.props || (current.props = {}))[attrName] = value;
            attrName = "";
        }
        else if (!inTag) {
            value && current.children.push(value);
        }
    }
    for (let i = 0; i < htmls.length; i++) {
        if (i)
            commit(i);
        let html = htmls[i];
        let last;
        let r;
        while (html) {
            /* istanbul ignore next */
            if (last === html)
                throw new Error("parse error:\n\t" + last);
            last = html;
            if (inComment) {
                if ((r = ~html.search(commentEndRegexp)))
                    inComment = false;
                html = html.slice(r ? ~r + 3 : html.length);
            }
            else if (inTag) {
                /* istanbul ignore else */
                if ((r = html.match(tagEndRegexp))) {
                    html = html.slice(r[0].length);
                    if (r[1] || voidTagNameRegexp.test(current.tag)) {
                        current = stack.pop();
                    }
                    inTag = false;
                }
                else if (attrName && (r = html.match(attributeValueRegexp))) {
                    html = html.slice(r[0].length);
                    commit(r[1] || r[2] || r[3] || "");
                }
                else if ((r = html.match(attributeNameRegexp))) {
                    html = html.slice(r[0].length).trim();
                    attrName = r[1];
                    if (r[1] !== "..." && !r[2])
                        commit("");
                }
            }
            else {
                if ((r = html.match(doctypeRegexp))) {
                    html = html.slice(r[0].length);
                }
                else if ((r = html.match(commentStartRegexp))) {
                    html = html.slice(2);
                    inComment = true;
                }
                else if ((r = html.match(closeTagRegexp))) {
                    html = html.slice(r[0].length);
                    if (current.tag !== r[1]) {
                        let j = stack.length;
                        while (j > 0 && stack[j - 1].tag !== r[1])
                            j--;
                        if (j) {
                            stack.length = j - 1;
                            current = stack.pop();
                        }
                    }
                    else {
                        current = stack.pop();
                    }
                }
                else if ((r = html.match(openTagRegexp))) {
                    html = html.slice(r[0].length);
                    stack.push(current);
                    current.children.push((current = { tag: r[1], children: [] }));
                    inTag = true;
                }
                else {
                    r = html.indexOf("<") || html.indexOf("<", 1);
                    const text = html.slice(0, ~r ? r : html.length);
                    html = html.slice(text.length);
                    commit(text.replace(/^\s*\n\s*|\s*\n\s*$/g, ""));
                }
            }
        }
    }
    return root.children;
}

const Template = Symbol();
function createTemplate(vdom, args) {
    return { [Template]: [vdom, args] };
}
function isTemplate(t) {
    return !!(t && t[Template]);
}
function isVNode(vdom) {
    return vdom && !!vdom.tag;
}
function items(t) {
    return t[Template];
}
function getArgs(t) {
    return items(t)[1];
}
function getKey(t) {
    var _a, _b;
    if (!isTemplate(t))
        return undefined;
    const [vdom, args] = items(t);
    /* istanbul ignore next */
    let key = isVNode(vdom[0]) && ((_b = (_a = vdom[0]) === null || _a === void 0 ? void 0 : _a.props) === null || _b === void 0 ? void 0 : _b[":key"]);
    if (typeof key === "number")
        key = args[key];
    return key;
}
function equalsTemplate(t1, t2) {
    return items(t1)[1][0] === items(t2)[1][0];
}
function isSvgTag(tag) {
    return tag.toLowerCase() === "svg";
}

function mount(vdom, parent, isSvg, mutations = []) {
    if (Array.isArray(vdom)) {
        vdom.forEach(v => mount(v, parent, isSvg, mutations));
    }
    else if (!isVNode(vdom)) {
        if (typeof vdom === "number") {
            mutations[vdom] = {
                node: parent.appendChild(document.createComment("")),
                isSvg
            };
        }
        else {
            parent.appendChild(document.createTextNode(vdom));
        }
    }
    else {
        const { tag, props, children } = vdom;
        isSvg = isSvgTag(tag) || isSvg;
        const node = parent.appendChild(isSvg
            ? document.createElementNS("http://www.w3.org/2000/svg", tag)
            : document.createElement(tag));
        props &&
            Object.keys(props).forEach(name => {
                if (typeof props[name] === "number") {
                    mutations[props[name]] = { node, name };
                }
                else {
                    node.setAttribute(name, props[name]);
                }
            });
        mount(children, node, isSvg, mutations);
    }
    return mutations;
}

function isAttributeMutation(m) {
    return !!m.name;
}
function isTemplateMutation(m) {
    return !!m._mutations;
}

function commit(mutations, args) {
    const q = [];
    for (let i = 1; i < mutations.length; i++) {
        const m = mutations[i];
        if (m) {
            if (isAttributeMutation(m) && m.name[0] === ".") {
                q.push(() => commitMutation(m, args[i]));
            }
            else {
                commitMutation(m, args[i]);
            }
        }
    }
    q.forEach(f => f());
}
function commitMutation(m, arg) {
    if (m._prev !== arg) {
        isAttributeMutation(m)
            ? commitAttribute(m.node, m.name, arg, m._prev)
            : commitNode(m, arg, m._prev, 0);
        m._prev = arg;
    }
}
function commitAttribute(node, name, next, prev) {
    if (name === "...") {
        if (typeof next === "object" && !Array.isArray(next) && next) {
            Object.keys(next).forEach(key => commitAttribute(node, key, next[key], prev && prev[key]));
        }
    }
    else if (name[0] === "?" && isRenderableValue(next)) {
        name = name.slice(1);
        next ? node.setAttribute(name, "") : node.removeAttribute(name);
    }
    else if (name[0] === ".") {
        node[name.slice(1)] = next;
    }
    else if (name[0] === "@") {
        name = name.slice(1);
        isEventListener(prev) && node.removeEventListener(name, prev);
        isEventListener(next) && node.addEventListener(name, next);
    }
    else if (name === ":key") ;
    else if (name === ":ref") {
        typeof next === "function"
            ? next(node)
            : (next.current = node);
    }
    else if (name === ":style") {
        typeof next === "object" &&
            setStyles(node.style, prev, next);
    }
    else if (name === ":class") {
        if (typeof next === "object") {
            const prevClassNames = classNames(prev);
            const nextClassNames = classNames(next);
            prevClassNames.forEach(p => nextClassNames.includes(p) || node.classList.remove(p));
            nextClassNames.forEach(n => prevClassNames.includes(n) || node.classList.add(n));
        }
    }
    else {
        isRenderableValue(next)
            ? node.setAttribute(name, next)
            : node.removeAttribute(name);
    }
}
function isRenderableValue(value) {
    return (value != null && typeof value !== "function" && typeof value !== "symbol");
}
function isEventListener(value) {
    return (typeof value === "function" ||
        (typeof value === "object" &&
            value != null &&
            "handleEvent" in value));
}
function setStyles(style, prev, next) {
    for (const i in prev)
        if (!next || !(i in next))
            style[i] = null;
    for (const i in next)
        if (!prev || prev[i] !== next[i])
            style[i] = next[i];
}
function classNames(value) {
    if (value == null)
        return [];
    if (Array.isArray(value))
        return value;
    return Object.keys(value).filter(i => !!value[i]);
}
function commitNode(mutation, next, prev, index) {
    next = isRenderableValue(next) ? next : null;
    prev = isRenderableValue(prev) ? prev : null;
    if (isIterable(next) || isIterable(prev)) {
        commitIterable(mutation, (next || []), (prev || []));
    }
    else {
        const tmpl = mutation;
        const txt = mutation;
        if (isTemplate(prev) && !isTemplate(next)) {
            removeTemplate(tmpl._mutations[index]);
            delete tmpl._mutations[index];
        }
        if (isText(prev) && !isText(next)) {
            removeNode(txt._texts[index]);
            delete txt._texts[index];
        }
        if (isTemplate(next) && !isTemplate(prev)) {
            (tmpl._mutations = tmpl._mutations || [])[index] = insertTemplate(
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            next, tmpl.node, tmpl.isSvg);
        }
        if (isText(next) && !isText(prev)) {
            insertNode(((txt._texts = txt._texts || [])[index] = document.createTextNode(next)), txt.node);
        }
        if (isTemplate(next) && isTemplate(prev)) {
            if (equalsTemplate(next, prev)) {
                commit(tmpl._mutations[index], getArgs(next));
            }
            else {
                removeTemplate(tmpl._mutations[index]);
                tmpl._mutations[index] = insertTemplate(next, tmpl.node, tmpl.isSvg);
            }
        }
        if (isText(next) && isText(prev)) {
            txt._texts[index].data = next;
        }
    }
}
function isIterable(value) {
    return (Array.isArray(value) ||
        !!(value &&
            typeof value !== "string" &&
            value[Symbol.iterator]));
}
function isText(t) {
    return t != null && !isTemplate(t);
}
function commitIterable(mutation, next, prev) {
    next = isIterable(next) ? next : [next];
    prev = isIterable(prev) ? prev : [prev];
    if (isTemplateMutation(mutation) && getKey(next[0]) != null) {
        commitTemplatesWithKey(mutation, next, prev);
    }
    else {
        for (let i = 0; i < prev.length || i < next.length; i++) {
            commitNode(mutation, next[i], prev[i], i);
        }
    }
}
function commitTemplatesWithKey(parentMutation, nexts, prevs) {
    const mutations = parentMutation._mutations;
    const nextMutations = [];
    const nextKeys = nexts.map(getKey);
    const prevKeys = prevs.map(getKey);
    let prevIndex = 0;
    let nextIndex = 0;
    while (prevIndex < prevKeys.length || nextIndex < nextKeys.length) {
        const next = nexts[nextIndex];
        if (nextKeys[nextIndex] != null &&
            !prevKeys.includes(nextKeys[nextIndex])) {
            nextMutations[nextIndex++] = insertTemplate(next, mutations[prevIndex]
                ? mutations[prevIndex]._marks[0]
                : parentMutation.node, parentMutation.isSvg);
        }
        else if (prevKeys[prevIndex] != null &&
            !nextKeys.includes(prevKeys[prevIndex])) {
            removeTemplate(mutations[prevIndex++]);
        }
        else {
            commit((nextMutations[nextIndex++] = mutations[prevIndex++]), getArgs(next));
        }
    }
    parentMutation._mutations = nextMutations;
}
function insertTemplate(template, refNode, isSvg) {
    const [vdom, args] = items(template);
    const fragment = document.createDocumentFragment();
    const mutations = Object.assign([], {
        _marks: [document.createComment(""), document.createComment("")]
    });
    mount(vdom, fragment, isSvg, mutations);
    commit(mutations, args);
    insertNode(mutations._marks[0], refNode);
    insertNode(fragment, refNode);
    insertNode(mutations._marks[1], refNode);
    return mutations;
}
function removeTemplate(mutations) {
    // eslint-disable-next-line prefer-const
    let [start, end] = mutations._marks;
    while (start !== end) {
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        const next = start.nextSibling;
        removeNode(start);
        start = next;
    }
    removeNode(end);
}
function insertNode(newNode, refNode) {
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    refNode.parentNode.insertBefore(newNode, refNode);
}
function removeNode(node) {
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    node.parentNode.removeChild(node);
}

const Cache = new WeakMap();
function html(strings, ..._) {
    let vdom = Cache.get(strings);
    vdom || Cache.set(strings, (vdom = parse(strings.raw)));
    /* eslint-disable-next-line prefer-rest-params */
    return createTemplate(vdom, arguments);
}
function render(template, container) {
    let mutations = Cache.get(container);
    mutations ||
        Cache.set(container, (mutations = mount(1, container, isSvgTag(container.nodeName))));
    commit(mutations, [, template]);
}

exports.html = html;
exports.isTemplate = isTemplate;
exports.isVNode = isVNode;
exports.items = items;
exports.render = render;
